# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2022 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

#===----------------------------------------------------------------------===//
#
# This file defines all of the Internal-specific intrinsics, which correspond to
# vISA instructions.
#
# Comment lines with a triple slash ### introduction are extracted and
# appended to docs/Targets/Internal/InternalLangRef.rst to give the Internal backend
# language reference in docs/autogenerated/Targets/Internal/InternalLangRef.rst.
#
#===------------------------------------------------------------------------===#

#------------ Currently Supported Types ----------------------
#PointerTypes = ["ptr_private", "ptr_global", "ptr_constant", "ptr_local", "ptr_generic"]
#FloatingPointTypes = ["half", "float", "double"]
#IntegerTypes = ["bool", "char", "short", "int", "long"]
#AdditionalTypes = ["vararg"]
#IntrinsicsProperties = ["None", "NoMem", "ReadArgMem", "ReadMem", "ReadWriteArgMem", "NoReturn", "NoDuplicate", "Convergent"]
#IntrinsicsProperties may be specified as a comma separated list(e.g., "Convergent,NoMem")
#
# EX. "blah": {"result" : {return_type}, "arguments" : [arg1_type, arg2_type.....], "attributes" : Property }
#
# The "any" type can be followed by a default type if a type is not explicitly specified : Ex. "any:int"
#
# 0 - LLVMMatchType<0>
# 1 - LLVMMatchType<1>
# {int} - LLVMMatchType<{int}>

Imported_Intrinsics = \
{

## ``llvm.vc.internal.jump.table`` : CMC internal, no VISA
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: integer BasicBlock index in the full set of destinations
## * arg1-N: the full set of switch labels
##
## * Return value: selected label
##
## The intrinsic is a helper for switch jump tables generation. Arg0
## will be used by visa switchjmp as index. Return value and arg1-N are
## used to make ir semantically legal.
##
    "jump_table" : { "result" : "anyptr",
                     "arguments" :  ["anyint", "vararg"],
                     "attributes" :  "NoMem"
                   },

## ``llvm.vc.internal.read.variable.region`` : read a vISA variable region
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: ptr pointer to a global variable that corresponds to a vISA variable
##         (overloaded)
## * arg1: i32 vstride in elements, constant
## * arg2: i32 width in elements, constant
## * arg3: i32 stride in elements, constant
## * arg4: i32 offset in elements, constant
##
## * Return value: iN, fN, vXiN, vXfN the read value (overloaded)
##
## This corresponds to MOV instruction or a general source operand in visa.
## Utilizes technique of using global variable in LLVM IR for predefined
## vISA variables.
##
    "read_variable_region" : { "result": "any",
                               "arguments" : ["anyptr", "int", "int", "int",
                                              "int"],
                               "attributes" : "ReadMem",
                              },

## ``llvm.vc.internal.write.variable.region`` : write a vISA variable region
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: ptr pointer to a global variable that corresponds to a vISA variable
##         (overloaded)
## * arg1: iN, fN, vXiN, vXfN value to write (overloaded)
## * arg2: i32 stride in elements, constant
## * arg3: i32 offset in elements, constant
## * arg4: i1 or vXi1 mask (overloaded)
##
## This corresponds to MOV instruction or a general destination operand in visa.
## Utilizes technique of using global variable in LLVM IR for predefined
## vISA variables.
##
    "write_variable_region" : { "result": "void",
                                "arguments" : ["anyptr", "any", "int",
                                               "int", "anyint"],
                                "attributes" : "WriteMem",
                              },

}



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import sys
from itertools import product

SignedDivSmall = {
    'types' : [
        { 'src_t' : 'int8_t', 'rem_t' : 'int8_t'},
        { 'src_t' : 'int16_t', 'rem_t' : 'int16_t'},
        { 'src_t' : 'int64_t', 'rem_t' : 'uint64_t'}
    ],
    'size' : [0] + [2 ** x for x in range(6)],
    'rounding' : ['rte'],
    'items' : ['__IREM', '__IDIV']
}
SignedDiv32 = {
    'types' : [
        { 'src_t' : 'int32_t', 'rem_t' : 'uint32_t'},
    ],
    'size' : [0] + [2 ** x for x in range(6)],
    'rounding' : ['rtz'],
    'items' : ['__IREM', '__IDIV']
}
UnsignedDiv = {
    'types' : [
        { 'src_t' : 'uint8_t', 'rem_t' : 'uint8_t'},
        { 'src_t' : 'uint16_t', 'rem_t' : 'uint16_t'},
        { 'src_t' : 'uint32_t', 'rem_t' : 'uint32_t'},
    ],
    'size' : [0] + [2 ** x for x in range(6)],
    'rounding' : ['rtz'],
    'items' : ['__UREM', '__UDIV']
}
Unsigned64Div = {
    'types' : [
        { 'src_t' : 'uint64_t', 'rem_t' : 'uint64_t'}
    ],
    'size' : [0] + [2 ** x for x in range(6)],
    'rounding' : ['rte'],
    'items' : ['__UREM', '__UDIV']
}

arg_num = len(sys.argv)
if arg_num > 1:
  output = sys.argv[1]
  f = open(output, 'w')
  sys.stdout = f

for desc in [SignedDivSmall, SignedDiv32, UnsignedDiv, Unsigned64Div]:
  for fname, size, rounding, types in product(desc['items'], desc['size'],
                                              desc['rounding'], desc['types']):
    if size == 0:
      print("{}_SCALAR_IMPL({}, {}, {}, {})".format(
        fname, types['src_t'], types['rem_t'], 1, rounding))
    else:
      print("{}_VECTOR_IMPL({}, {}, {}, {})".format(
        fname, types['src_t'], types['rem_t'], size, rounding))



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2017-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import os
import sys
import errno
import re
from typing import List


class DeclHeader:
    # line contains the entire string of the line the decl was found on
    line: str
    # declName is just the identifier name
    declName: str
    # fields contains a list of all the names of the fields in the structure
    fields: List[str]

    def __init__(self, line_p, declName_p, fields_p):
        self.line = line_p
        self.declName = declName_p
        self.fields = fields_p

enumNames: List[DeclHeader] = []
structureNames: List[DeclHeader] = []


def parseCmdArgs():
    if (len(sys.argv) != 3):
        sys.exit("usage: autogen.py <path_to_MDFrameWork.h> <path_to_MDNodeFuncs.gen>")

    # usage: autogen.py <path_to_MDFrameWork.h> <path_to_MDNodeFuncs.gen>
    __MDFrameWorkFile__ = sys.argv[1]
    __genFile__         = sys.argv[2]

    if not os.path.isfile(__MDFrameWorkFile__):
        sys.exit("Could not find the file " + __MDFrameWorkFile__)

    __genDir__ = os.path.dirname(__genFile__)
    # Try to create the path to MDNodeFuncs.gen if it doesn't already exist.
    if not os.path.exists(__genDir__):
        try:
            os.makedirs(__genDir__)
        except OSError as err:
            # In case of a failure to create this directory, and the directory doesn't already exist.
            if err.errno != errno.EEXIST:
                sys.exit("Failed to create the directory " + __genDir__)

    return __MDFrameWorkFile__ , __genFile__


def storeVars(line, declHeader):
    vars = line.split()
    for i in range(1, len(vars)):
        res = vars[i].split(",")
        for j in range(0, len(res)):
            if(res[j] != '' and res[j]!= ';' and res[j].find("ModuleMetaData") == -1 and res[j].find("FunctionMetaData") == -1):
                if(i != len(vars)-1):
                    declHeader.fields.append(res[j] + ';')
                else:
                    if(j != len(res)-1):
                        declHeader.fields.append(res[j] + ";")
                    else:
                        declHeader.fields.append(res[j])


def extractVars(line, declHeader):
    vars = line.split()
    if any(ty in line for ty in ("std::vector", "std::map", "std::array", "MapVector")):
        declHeader.fields.append(vars[len(vars)-1])
        return
    if line.find("=") != -1:
        declHeader.fields.append(vars[vars.index("=")-1]+";")
    elif(line.find(",")) == -1:
        if len(vars) == 2:
            declHeader.fields.append(vars[1])
        else:
            if(len(vars) == 3):
                declHeader.fields.append(vars[len(vars)-1])
    else:
        storeVars(line, declHeader)


def extractEnumVal(line, declHeader):
    vars = line.split()
    if (len(vars) == 0) or (line.find("{") != -1):
        return

    val = vars[0]
    if val[-1] == ',':
        val = val[:-1]

    declHeader.fields.append(val)


def parseFile(fileName, insideIGCNameSpace):
    inputFile = None
    try:
        inputFile = open(fileName, 'r')
    except:
        sys.exit("Failed to open the file " + fileName)

    # with statement automatically closes the file
    with inputFile as file:
        pcount = 0
        for line in file:
            line = line.split("//")[0]

            if line.find("namespace IGC") != -1:
                while line.find("{") == -1:
                    line = next(file, None)
                insideIGCNameSpace = True
                pcount += 1
                #
            if insideIGCNameSpace:
                blockType = re.search("struct|enum", line)
                if blockType:
                    words = line.split()
                    idx = 2 if 'class' in words else 1
                    foundDecl = DeclHeader(line, words[idx], [])
                    opcount = pcount
                    namesList = structureNames
                    extractFunc = extractVars
                    if blockType[0] == 'enum':
                        namesList = enumNames
                        extractFunc = extractEnumVal
                    while True:
                        line = next(file, None)
                        assert line, "EOF reached with unclosed enum or struct, check " + fileName + " formatting"
                        assert re.search("/\*", line) == None, "Multi-line comments are not supported yet"
                        line = line.split("//")[0]
                        pcount += line.count("{") - line.count("}")
                        if pcount <= opcount:
                            break
                        extractFunc(re.sub("{|}","", line), foundDecl)
                    assert pcount == opcount, "Unexpected struct/enum ending, check " + fileName + " formatting"
                    namesList.append(foundDecl)
                elif line.lstrip().startswith("#include") == True:
                    words = line.split()
                    parent_dir = os.path.dirname(fileName)
                    include_file = words[1][1:-1]  # cut off the "" or <> surrounding the file name
                    include_file_path = os.path.join(parent_dir, include_file)
                    parseFile(include_file_path, True)
                elif line.find("}") != -1 and line.find("};") == -1:
                    insideIGCNameSpace = False
                    pcount -= 1
        assert pcount == 0, "EOF reached, with unclosed IGC namespace, check " + fileName + " formatting"

def printStructCalls(structDecl, outputFile):
    outputFile.write("    Metadata* v[] = \n")
    outputFile.write("    { \n")
    outputFile.write("        MDString::get(module->getContext(), name),\n")
    for item in structDecl.fields:
        item = item[:-1]
        p =  '"{}"'.format(item)
        outputFile.write("        CreateNode(" + structDecl.declName + "Var" + "." + item + ", module, ")
        outputFile.write(p)
        outputFile.write("),\n")
    outputFile.write("    };\n")
    outputFile.write("    MDNode* node = MDNode::get(module->getContext(), v);\n")
    outputFile.write("    return node;\n")


def printEnumCalls(enumDecl, outputFile):
    outputFile.write("    StringRef enumName;\n")
    outputFile.write("    switch("+ enumDecl.declName + "Var)\n")
    outputFile.write("    {\n")
    for item in enumDecl.fields:
        outputFile.write("        case IGC::" + enumDecl.declName + "::" + item + ":\n"  )
        outputFile.write("            enumName = ")
        outputFile.write("\"" + item + "\"")
        outputFile.write(";\n")
        outputFile.write("            break;\n" )
    outputFile.write("    }\n")
    outputFile.write("    Metadata* v[] = \n")
    outputFile.write("    { \n")
    outputFile.write("        MDString::get(module->getContext(), name),\n")
    outputFile.write("        MDString::get(module->getContext(), enumName),\n")
    outputFile.write("    };\n")
    outputFile.write("    MDNode* node = MDNode::get(module->getContext(), v);\n")
    outputFile.write("    return node;\n")


def printStructReadCalls(structDecl, outputFile):
     for item in structDecl.fields:
        item = item[:-1]
        p =  '"{}"'.format(item)
        outputFile.write("    readNode(" + structDecl.declName + "Var" + "." + item + ", node , ")
        outputFile.write(p)
        outputFile.write(");\n")

def printEnumReadCalls(enumDecl, outputFile):
    outputFile.write("    StringRef s = cast<MDString>(node->getOperand(1))->getString();\n")
    outputFile.write("    std::string str = s.str();\n")
    outputFile.write("    "+ enumDecl.declName + "Var = (IGC::" + enumDecl.declName + ")(0);\n")

    for item in enumDecl.fields:
        outputFile.write("    if((str.size() == sizeof(\""+ item + "\")-1) && (::memcmp(str.c_str(),")
        outputFile.write("\"" + item + "\"")
        outputFile.write(",str.size())==0))\n")
        outputFile.write("    {\n")
        outputFile.write("            "+ enumDecl.declName + "Var = IGC::" + enumDecl.declName + "::" + item + ";\n")
        outputFile.write("    } else\n")

    outputFile.write("    {\n")
    outputFile.write("            "+ enumDecl.declName + "Var = (IGC::" + enumDecl.declName + ")(0);\n")
    outputFile.write("    }\n")


def emitCodeBlock(names: List[DeclHeader], declType, fmtFn, printFn, outputFile):
    for item in names:
        outputFile.write(fmtFn(item.declName))
        outputFile.write("{\n")
        printFn(item, outputFile)
        outputFile.write("}\n\n")


def emitEnumCreateNode(outputFile):
    def fmtFn(item):
        return f"MDNode* CreateNode(IGC::{item} {item}Var, Module* module, StringRef name)\n"
    emitCodeBlock(enumNames, "enum", fmtFn, printEnumCalls, outputFile)

def emitStructCreateNode(outputFile):
    def fmtFn(item):
        return "MDNode* CreateNode(const IGC::" + item + "& " + item + "Var" +", Module* module, StringRef name)\n"
    emitCodeBlock(structureNames, "struct", fmtFn, printStructCalls, outputFile)

def emitEnumReadNode(outputFile):
    def fmtFn(item):
        return "void readNode( IGC::" + item + " &" + item + "Var," + " MDNode* node)\n"
    emitCodeBlock(enumNames, "enum", fmtFn, printEnumReadCalls, outputFile)

def emitStructReadNode(outputFile):
    def fmtFn(item):
        return "void readNode( IGC::" + item + " &" + item + "Var," + " MDNode* node)\n"
    emitCodeBlock(structureNames, "struct", fmtFn, printStructReadCalls, outputFile)


def genCode(fileName):
    outputFile = None
    try:
        outputFile = open(fileName, 'w')
    except:
        sys.exit("Failed to open the file " + fileName)

    emitEnumCreateNode(outputFile)
    emitStructCreateNode(outputFile)
    emitEnumReadNode(outputFile)
    emitStructReadNode(outputFile)

    outputFile.close()


if __name__ == '__main__':
    __MDFrameWorkFile__ , __genFile__ = parseCmdArgs()
    parseFile(__MDFrameWorkFile__, False)
    genCode(__genFile__)
